# vim: set fileencoding=utf-8 :

# connord - connect to nordvpn servers
# Copyright (C) 2019  Mael Stor <maelstor@posteo.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import time
import requests
from progress.bar import IncrementalBar
from cachetools import cached, LRUCache, TTLCache
from connord import ConnordError
from connord import servers
from connord import sqlite
from connord.formatter import Formatter


class AreaError(ConnordError):
    """Thrown within this module"""


API_URL = "https://nominatim.openstreetmap.org"


@cached(cache=LRUCache(maxsize=50))
def query_location(latitude, longitude):
    header = {
        "User-Agent": "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:60.0) "
        "Gecko/20100101 Firefox/60.0"
    }

    endpoint = "reverse"
    flags = {
        "lat": latitude,
        "lon": longitude,
        "format": "jsonv2",
        "addressdetails": "1",
        "accept-language": "en",
        "zoom": "18",
    }
    url = "{}/{}.php?".format(API_URL, endpoint)
    for k, v in flags.items():
        url += "{}={}&".format(k, v)

    url = url.rstrip("&")
    with requests.get(url, headers=header, timeout=1) as response:
        time.sleep(1)
        return response.json()


def init_database(connection):
    with connection:
        sqlite.create_location_table(connection)


def update_database():
    connection = sqlite.create_connection()
    with connection:
        init_database(connection)

    servers_ = servers.get_servers()
    for server in IncrementalBar("Updating locations", max=len(servers_)).iter(
        servers_
    ):

        longitude = str(server["location"]["long"])
        latitude = str(server["location"]["lat"])

        connection = sqlite.create_connection()
        with connection:
            if sqlite.location_exists(connection, latitude, longitude):
                continue
            else:
                location_d = query_location(latitude, longitude)

        display_name = location_d["display_name"]
        city_keys = ["city", "town", "village", "residential", "state"]
        for key in city_keys:
            try:
                city = location_d["address"][key]
                break
            except KeyError:
                continue
        else:
            city = "Unknown"

        country = server["country"]
        country_code = server["flag"].lower()
        location = (latitude, longitude, display_name, city, country, country_code)

        connection = sqlite.create_connection()
        with connection:
            sqlite.create_location(connection, location)


def get_server_angulars(server):
    latitude = str(server["location"]["lat"])
    longitude = str(server["location"]["long"])
    return (latitude, longitude)


def verify_areas(areas_):
    if not isinstance(areas_, list):
        raise AreaError("Wrong areas: {!s}".format(areas_))

    locations = get_locations()

    areas_not_found = []
    # side effect: get rid of double entries in areas_ from command-line
    areas_found = {area: list() for area in areas_}
    for area in areas_found.keys():
        for location in locations:
            city = location["city"]
            city_lower = city.lower()
            if city_lower.startswith(area):
                areas_found[area].append(city)

        if not areas_found[area]:
            areas_not_found.append(area)

    if areas_not_found:
        raise ValueError("Areas not found: {!s}".format(areas_not_found))

    ambigous_areas = {}
    for area, cities in areas_found.items():
        if len(cities) > 1:
            ambigous_areas[area] = cities

    if ambigous_areas:
        error_string = ""
        for area, cities in ambigous_areas.items():
            error_string += " {}: {},".format(area, cities)

        error_string = error_string.rstrip(",")
        raise AreaError("Ambigous Areas:{}".format(error_string))

    return [area for area in areas_found.keys()]


def get_translation_table():
    return str.maketrans("áãčëéşșť", "aaceesst")


def filter_servers(servers_, areas_):
    """Filter servers by areas"""

    if servers_ is None:
        raise TypeError("Servers may not be None")

    if areas_ is None or not areas_ or not servers_:
        return servers_

    areas_lower = [str.lower(area) for area in areas_]
    areas_found = verify_areas(areas_lower)

    filtered_servers = []
    connection = sqlite.create_connection()
    with connection:
        translation_table = get_translation_table()
        for server in servers_:
            lat, lon = get_server_angulars(server)
            if not sqlite.location_exists(connection, lat, lon):
                update_database()

            city = sqlite.get_city(connection, lat, lon)[0]
            city = city.translate(translation_table)
            city = city.lower()
            for area in areas_found:
                if city.startswith(area):
                    filtered_servers.append(server)
                    break

    return filtered_servers


def get_min_id(city):
    min_id = ""
    word = ""
    translation_table = get_translation_table()
    for char in city:
        word += char.lower()
        char = char.translate(translation_table)
        min_id += char.lower()
        try:
            verify_areas([word])
            break
        except AreaError:
            continue
        except ValueError:
            continue

    return min_id


@cached(cache=TTLCache(ttl=60, maxsize=1))
def get_locations():
    connection = sqlite.create_connection()
    with connection:
        locations = sqlite.get_locations(connection)
        if not locations:
            update_database()
            return sqlite.get_locations(connection)

        return locations


class AreasPrettyFormatter(Formatter):
    def format_headline(self, sep="="):
        headline = self.format_ruler(sep) + "\n"
        headline += "{:8}: {:^15} {:^15}  {:40}\n".format(
            "Mini ID", "Latitude", "Longitude", "City"
        )
        headline += "{}\n".format("Address")
        headline += self.format_ruler(sep)
        return headline

    def format_area(self, location):
        lat = float(location["latitude"])
        lon = float(location["longitude"])
        display_name = location["display_name"]
        city = location["city"]
        min_id = get_min_id(city)

        string = "{!r:8}: {: 14.9f}° {: 14.9f}°  {:40}\n".format(min_id, lat, lon, city)
        string += "{}\n".format(display_name)
        string += self.format_ruler(sep="-")
        return string


def to_string(stream=False):
    formatter = AreasPrettyFormatter()
    file_ = formatter.get_stream_file(stream)

    locations = get_locations()

    headline = formatter.format_headline()
    print(headline, file=file_)

    for location in locations:
        area = formatter.format_area(location)
        print(area, file=file_)

    return formatter.get_output()


def print_areas():
    to_string(stream=True)



# vim: set fileencoding=utf-8 :

# connord - connect to nordvpn servers
# Copyright (C) 2019  Mael Stor <maelstor@posteo.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import subprocess
from multiprocessing.pool import ThreadPool
from multiprocessing import cpu_count
from math import inf
import time
import os
import re
import signal
import tempfile
from connord import ConnordError
from connord import iptables
from connord import servers
from connord import load
from connord import countries
from connord import areas
from connord import types
from connord import features
from connord import user
from connord import resources
from connord import update


class ConnectError(ConnordError):
    """Thrown within this module"""


def ping(server):
    """
    Ping a server
    :param dict server: A server as dictionary
    :returns: copy of server with additional key 'ping'
    """

    server_copy = server.copy()
    pattern = re.compile(r"rtt .* = ([\d\.]+)/([\d\.]+)/([\d\.]+)/.* ms")
    ip_address = server["ip_address"]
    with subprocess.Popen(
        ["ping", "-q", "-n", "-c", "1", "-l", "1", "-W", "1", ip_address],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    ) as ping_:

        out, _ = ping_.communicate()
        mat = pattern.search(out.decode())
        if mat:
            server_copy["ping"] = float(mat.group(2))
        else:
            server_copy["ping"] = float("inf")

        return server_copy


def ping_servers_parallel(servers_):
    """
    Ping a list of servers
    :param list servers_: List of servers
    :returns: List of servers with ping time
    """
    worker_count = cpu_count() + 1
    with ThreadPool(processes=worker_count) as pool:
        results = []
        for server in servers_:
            results.append(pool.apply_async(ping, (server,)))

        pinged_servers = []
        for result in results:
            pinged_servers.append(result.get())

        return pinged_servers


# pylint: disable=too-many-arguments
def filter_servers(
    servers_, netflix, countries_, areas_, features_, types_, load_, match
):
    servers_ = servers_.copy()
    if load_:
        servers_ = load.filter_servers(servers_, load_, match)
    if netflix:
        servers_ = servers.filter_netflix_servers(servers_, countries_)
    if countries_:
        servers_ = countries.filter_servers(servers_, countries_)
    if areas_:
        servers_ = areas.filter_servers(servers_, areas_)
    if types_:
        servers_ = types.filter_servers(servers_, types_)
    if features_:
        servers_ = features.filter_servers(servers_, features_)

    return servers_


def filter_best_servers(servers_):
    servers_ = servers_.copy()
    servers_ = sorted(servers_, key=lambda k: k["load"])
    if len(servers_) > 10:
        servers_ = servers_[:10]
    servers_ = ping_servers_parallel(servers_)
    servers_ = sorted(servers_, key=lambda k: k["ping"])
    return servers_


@user.needs_root
def connect_to_specific_server(domain, openvpn, daemon, protocol):
    server = servers.get_server_by_domain(domain[0])
    return run_openvpn(server, openvpn, daemon, protocol)


# pylint: disable=too-many-locals
@user.needs_root
def connect(
    domain,
    countries_,
    areas_,
    features_,
    types_,
    netflix,
    load_,
    match,
    daemon,
    config_,
    openvpn,
    protocol,
):

    if domain != "best":
        return connect_to_specific_server(domain, openvpn, daemon, protocol)

    if protocol:
        feature = "openvpn_" + protocol
        if features_ is None:
            features_ = [feature]
        elif feature in features_:
            pass
        else:
            features_.append("openvpn" + protocol)

    servers_ = servers.get_servers()
    servers_ = filter_servers(
        servers_, netflix, countries_, areas_, features_, types_, load_, match
    )

    best_servers = filter_best_servers(servers_)
    max_retries = 3
    for i, server in enumerate(best_servers):
        if i == max_retries:
            raise ConnectError("Maximum retries reached.")

        if server["ping"] != inf:
            if run_openvpn(server, openvpn, daemon, protocol):
                return True
            # else give the next server a try
        else:
            raise ConnectError("No server left with a valid ping.")

    raise ConnectError("No server found to establish a connection.")


class OpenvpnCommandPanic(ConnectError):
    def __init__(self, problem, message=None):
        if not message:
            message = "Running openvpn failed: {}".format(problem)

        super().__init__(message)
        self.problem = problem


class OpenvpnCommand:
    def __init__(self, server, openvpn, daemon, protocol):
        self.server = server
        self.domain = server["domain"]
        self.openvpn_options = openvpn
        self.daemon = daemon
        self.protocol = protocol
        self.cmd = ["/usr/bin/openvpn"]

    def _add_openvpn_cmd_option(self, flag, *args):
        if flag.startswith("-"):
            if flag not in self.cmd:
                self.cmd.append(flag)
                for arg in args:
                    self.cmd.append(arg)
        else:  # flag seems to be an argument and can be added without checks
            self.cmd.append(flag)

    @staticmethod
    def _format_flag(flag):
        return "--{}".format(flag)

    def _forge_number(self, key, value):
        flag = self._format_flag(key)
        self._add_openvpn_cmd_option(flag, str(value))

    def _forge_string(self, key, value):
        flag = self._format_flag(key)
        if key == "auth-user-pass":
            if value == "built-in":
                creds_file = resources.get_credentials_file()
            else:
                creds_file = value

            self._add_openvpn_cmd_option(flag, creds_file)
        else:
            self._add_openvpn_cmd_option(flag, value)

    def _forge_bool(self, key, value):
        if value and value not in ("False", "false"):
            flag = self._format_flag(key)
            self._add_openvpn_cmd_option(flag)

    def _forge_list(self, key, list_):
        if key == "scripts":
            self._forge_scripts(list_)
        else:
            flag = self._format_flag(key)
            self._add_openvpn_cmd_option(flag, list_)

    @staticmethod
    def _format_script_arg(script_name, path, file_):
        if path == "built-in":
            script_path = resources.get_scripts_file(script_name=script_name)
            env_dir = resources.get_stats_dir()
            env_file = "{}/{}".format(env_dir, file_)
        else:
            script_path = path
            env_file = file_

        return "{} {}".format(script_path, env_file)

    def _forge_scripts(self, scripts):
        for script in scripts:
            name = script["name"]
            flag = self._format_flag(name)
            path = script["path"]
            file_ = script["creates"]
            if name in ("up", "down"):
                arg = self._format_script_arg("openvpn_up_down.bash", path, file_)
            elif name == "ipchange":
                arg = self._format_script_arg("openvpn_ipchange.bash", path, file_)
            else:
                if path == "built-in":
                    raise resources.ResourceNotFoundError(
                        "No built-in found for {!r}.".format(name)
                    )

                arg = self._format_script_arg(name, path, file_)

            self._add_openvpn_cmd_option(flag, arg)

    def _forge_command_line(self):
        if self.openvpn_options:
            for option in self.openvpn_options.split():
                self._add_openvpn_cmd_option(option)

        if self.daemon:
            self._add_openvpn_cmd_option("--daemon")

    def _forge_config(self):
        openvpn_config = resources.get_config()["openvpn"]
        for k, v in openvpn_config.items():
            if isinstance(v, bool) or v in ("true", "True", "false", "False"):
                self._forge_bool(k, v)
            elif isinstance(v, list):
                self._forge_list(k, v)
            elif isinstance(v, (int, float)):
                self._forge_number(k, v)
            else:
                self._forge_string(k, v)

    def has_flag(self, flag):
        if flag.startswith("--"):
            return flag in self.cmd

        return "--{}".format(flag) in self.cmd

    def forge_ovpn_config(self, config_file=None):
        if not config_file:
            config_file = resources.get_ovpn_config(self.domain, self.protocol)

        config_tmp = resources.get_ovpn_tmp_path()

        with open(config_file, "r") as config_fd:
            lines = config_fd.readlines()
        with open(config_tmp, "w") as config_fd:
            for line in lines:
                if line != "\n":
                    fake_line = line.rstrip() + " $"
                    flag, _ = fake_line.split(maxsplit=1)
                    if not self.has_flag(flag):
                        config_fd.write(line)
                else:
                    config_fd.write(line)

        os.chmod(config_tmp, mode=0o640)
        self._add_openvpn_cmd_option("--config", config_tmp)

    def forge(self):
        self._forge_command_line()
        self._forge_config()
        if "--config" not in self.cmd:
            try:
                self.forge_ovpn_config()
            except resources.ResourceNotFoundError:
                update.update(
                    force=True
                )  # give updating a try else let the error happen
                self.forge_ovpn_config()
        else:
            index = self.cmd.index("--config")
            config_file = self.cmd[index + 1]
            self.cmd.remove("--config")
            self.cmd.remove(config_file)
            self.forge_ovpn_config(config_file=config_file)

        if not self.has_flag("--writepid"):
            pid_dir = resources.get_stats_dir(create=True)
            pid_file = pid_dir + "/openvpn.pid"
            self._add_openvpn_cmd_option("--writepid", pid_file)

    def is_daemon(self):
        return "--daemon" in self.cmd

    @staticmethod
    def cleanup():
        resources.remove_stats_dir()
        iptables.reset(fallback=True)

    @staticmethod
    def is_running(process):
        return not bool(process.poll())

    def panic(self, process, problem):
        if self.is_running(process):
            process.kill()

        self.cleanup()
        resources.remove_ovpn_tmp_file()
        raise OpenvpnCommandPanic(problem)

    def run(self):
        config_dict = resources.get_config()["openvpn"]
        self.cleanup()

        with subprocess.Popen(self.cmd) as ovpn:
            # give openvpn a maximum of 60 seconds to startup. A lower value is bad if
            # asked for username/password.
            # pylint: disable=unused-variable
            for i in range(300):
                try:
                    if self.is_running(ovpn):
                        # delay initialization of iptables until resource files are
                        # created. If none are created the delay still applies as normal
                        # timeout
                        time.sleep(0.2)
                        for script in config_dict["scripts"]:
                            stage = script["stage"]
                            if stage in ("up", "always"):
                                resources.get_stats_file(
                                    stats_name=script["creates"], create=False
                                )

                    else:
                        self.panic(ovpn, "Openvpn process stopped unexpected.")

                    if iptables.apply_config_dir(self.server, self.protocol):
                        resources.write_stats(self.server, stats_name="server")

                        stats_dict = resources.get_stats()
                        stats_dict["last_server"] = {}
                        stats_dict["last_server"]["domain"] = self.domain
                        stats_dict["last_server"]["protocol"] = self.protocol
                        resources.write_stats(stats_dict)
                    else:
                        self.panic(ovpn, "Applying iptables failed.")

                    break
                except resources.ResourceNotFoundError:
                    pass
            ### for
            else:
                self.panic(ovpn, "Timeout reached.")

            if self.is_running(ovpn):
                ovpn.wait()

        return True


@user.needs_root
def run_openvpn(server, openvpn, daemon, protocol):
    openvpn_cmd = OpenvpnCommand(server, openvpn, daemon, protocol)
    openvpn_cmd.forge()

    retval = False
    try:
        retval = openvpn_cmd.run()
    except KeyboardInterrupt:
        retval = True
        time.sleep(1)

    if not openvpn_cmd.is_daemon():
        resources.remove_ovpn_tmp_file()
        openvpn_cmd.cleanup()

    return retval


@user.needs_root
def kill_openvpn(pid=None):
    if pid:
        os.kill(pid, signal.SIGTERM)
    else:
        cmd = ["ps"]
        cmd.append("-A")
        with subprocess.Popen(cmd, stdout=subprocess.PIPE) as proc:
            out, _ = proc.communicate()
            for line in out.decode().splitlines():
                if "openvpn" in line:
                    pid = int(line.split(None, 1)[0])
                    os.kill(pid, signal.SIGTERM)



# vim: set fileencoding=utf-8 :

# connord - connect to nordvpn servers
# Copyright (C) 2019  Mael Stor <maelstor@posteo.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Module to check root access and acquire it if needed"""

import os
import functools
import subprocess
from connord import ConnordError


class UserError(ConnordError):
    """Thrown within this module"""


def is_root():
    """
    Returns true if user is root

    :return: True if effective uid is 0
    """
    return os.geteuid() == 0


def needs_root(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if is_root():
            return func(*args, **kwargs)
        else:
            raise PermissionError(
                "Function '{}' needs root access.".format(func.__name__)
            )

    return wrapper


def prompt_sudo():
    """
    Prompt for sudo password and change effective uid to root on success.

    :return: True if user has root access now
    """

    if os.geteuid() != 0:
        message = "[connord][sudo] password needed: "
        return subprocess.check_call("sudo -v -p '{}'".format(message), shell=True)

    return True


# TODO: check_root throws PermissionError



# vim: set fileencoding=utf-8 :

# connord - connect to nordvpn servers
# Copyright (C) 2019  Mael Stor <maelstor@posteo.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Manages server features"""

from connord import ConnordError

FEATURES = {
    "ikev2": "IKEv2/IPSec Protocol",
    "openvpn_udp": "UDP",
    "openvpn_tcp": "TCP",
    "socks": "Socks 5",
    "proxy": "HTTP Proxy",
    "pptp": "PPTP",
    "l2tp": "L2TP/IPSec",
    "openvpn_xor_udp": "OpenVPN UDP Obfuscated",
    "openvpn_xor_tcp": "OpenVPN TCP Obfuscated",
    "proxy_cybersec": "HTTP Proxy CyberSec",
    "proxy_ssl": "HTTP Proxy (SSL)",
    "proxy_ssl_cybersec": "HTTP CyberSec Proxy (SSL)",
    "ikev2_v6": "IKEv2/IPSec IPv6",
    "openvpn_udp_v6": "UDPv6",
    "openvpn_tcp_v6": "TCPv6",
    "wireguard_udp": "WireGuard UDP",
    "openvpn_udp_tls_crypt": "UDP TLS encryption",
    "openvpn_tcp_tls_crypt": "TCP TLS encryption",
}


class FeatureError(ConnordError):
    """
    Thrown within this module
    """


def verify_features(features):
    """
    Verify if features are valid.

    :param features: List of features.
    :returns: True if all features are valid or else raises an exception.
    :raises FeatureError: raises a FeatureError if the format of the features
                          were invalid

    """

    if not isinstance(features, list):
        raise FeatureError("Wrong server features: {!s}".format(features))

    wrong_features = []
    for feature in features:
        if feature not in FEATURES.keys():
            wrong_features.append(feature)

    if wrong_features:
        raise FeatureError("Wrong server features: {!s}".format(wrong_features))

    return True


def filter_servers(servers, features=None):
    """
    Filter a list of servers by feature.

    :param servers: List of servers as parsed from nordvpn api in json format
    :param features: List of features to filter the server list. If None or
                     empty the default 'openvpn_udp' is applied.
    :returns: The filtered list
    """

    if features is None or not features:
        features = ["openvpn_udp"]

    verify_features(features)

    filtered_servers = []
    servers = servers.copy()
    for server in servers:
        append = True
        for feature in features:
            if not server["features"][feature]:
                append = False
                break

        if append:
            filtered_servers.append(server)

    return filtered_servers


def to_string():
    """
    Assemble all features as printable string

    :returns: A simple formatted string to use as output on screen
    """

    result = ""
    for feature, description in FEATURES.items():
        result += "{:26}{}\n".format(feature, description)

    return result.rstrip()


def print_features():
    """Prints all possible features"""
    print(to_string())



# vim: set fileencoding=utf-8 :

# connord - connect to nordvpn servers
# Copyright (C) 2019  Mael Stor <maelstor@posteo.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""List servers"""

from connord import servers
from connord import countries
from connord import types
from connord import features
from connord import load
from connord import iptables
from connord import areas
from connord.formatter import Formatter


def filter_servers_by_count(servers_, top):
    """
    Filter servers to just show top count results

    :param servers_: List of servers
    :param top: Integer to show count servers
    :returns: The filtered servers
    """

    servers_ = servers_.copy()
    filtered_servers = [server for i, server in enumerate(servers_, 1) if i <= top]
    return filtered_servers


def list_iptables(tables, version):
    iptables.print_iptables(tables, version)
    return True


class OverviewPrettyFormatter(Formatter):
    def __init__(self, output=None, max_line_length=80, stream=False):
        super().__init__(output, max_line_length)
        self.stream = stream
        self.has_output = False

    def print_countries(self, countries_):
        if countries_ and None in countries_:
            self.has_output = True
            file_ = self.get_stream_file(self.stream)

            country_header = "Countries"
            print(self.center_string(country_header, sep="="), file=file_)
            for country_code, country in countries.COUNTRIES.items():
                print("  {:6}{}".format(country_code, country), file=file_)
            print(self.format_ruler(sep="-"), file=file_)

    def print_types(self, types_):
        if types_ and None in types_:
            self.has_output = True
            file_ = self.get_stream_file(self.stream)

            types_header = self.center_string("Server Types", sep="=")
            print(types_header, file=file_)
            for server_type, description in types.TYPES.items():
                print("  {:26}{}".format(server_type, description), file=file_)
            print(self.format_ruler(sep="-"), file=file_)

    def print_features(self, features_):
        if features_ and None in features_:
            self.has_output = True
            file_ = self.get_stream_file(self.stream)

            features_header = "Server Features"
            print(self.center_string(features_header, sep="="), file=file_)
            for feature, description in features.FEATURES.items():
                print("  {:26}{}".format(feature, description), file=file_)
            print(self.format_ruler(sep="-"), file=file_)

    def print_areas(self, areas_):
        if areas_ and None in areas_:
            self.has_output = True
            areas.print_areas()


def filter_servers(
    servers_, netflix, countries_, areas_, features_, types_, load_, match, top
):
    servers_ = servers_.copy()
    if load_:
        servers_ = load.filter_servers(servers_, load_, match)
    if netflix:
        servers_ = servers.filter_netflix_servers(servers_, countries_)
    if countries_:
        servers_ = countries.filter_servers(servers_, countries_)
    if areas_:
        servers_ = areas.filter_servers(servers_, areas_)
    if types_:
        servers_ = types.filter_servers(servers_, types_)
    if features_:
        servers_ = features.filter_servers(servers_, features_)
    if top:
        servers_ = filter_servers_by_count(servers_, top)

    return servers_


# TODO: rename to list_servers
def main(countries_, areas_, types_, features_, netflix, load_, match, top):
    """
    Main method to do the actual listing and prints the resulting
    list of servers.

    :param countries_: List of countries
    :param area_: List of areas
    :param types_: List of types
    :param features_: List of features
    :param netflix: If set filter servers optimised for netflix
    :param load_: An integer to filter servers by load.
    :param match: Apply load filter with 'max', 'min' or 'exact' match
    :param top: Show just top count results.
    :returns: True
    """

    formatter = OverviewPrettyFormatter(stream=True)
    formatter.print_countries(countries_)
    formatter.print_areas(areas_)
    formatter.print_types(types_)
    formatter.print_features(features_)

    if not formatter.has_output:
        servers_ = servers.get_servers()
        servers_ = filter_servers(
            servers_,
            netflix,
            countries_,
            areas_,
            features_,
            types_,
            load_,
            match,
            top,
        )

        servers.to_string(servers_, stream=True)

    return True
